%{
  #include <stdio.h>
  #include "defs.h"
  #include "symtab.h"
  #include "codegen.h"

  bool check_types(int first_index, int second_index);

  extern int yylineno;
  char char_buffer[CHAR_BUFFER_LENGTH];
  int error_count = 0;
  int warning_count = 0;
  int var_num = 0;
  int fun_idx = -1;
  int fcall_idx = -1;
  int lab_num = -1;
  FILE *output;
%}

%union {
  int i;
  char *s;
}


%token <i> _TYPE
%token _IF
%token _ELSE
%token _RETURN
%token <s> _ID
%token <s> _INT_NUMBER
%token <s> _UINT_NUMBER
%token _LPAREN
%token _RPAREN
%token _LBRACKET
%token _RBRACKET
%token _ASSIGN
%token _SEMICOLON
%token <i> _AROP
%token <i> _RELOP

%type <i> type num_exp exp literal parameter
%type <i> function_call argument rel_exp

%nonassoc ONLY_IF
%nonassoc _ELSE

%%

program
  : function_list
      {   
        int idx;
        if((idx = lookup_symbol("main", FUN)) == -1)
          err("undefined reference to 'main'");
        else 
          if(get_type(idx) != INT)
            warn("return type of 'main' is not int"); 
      }
  ;

function_list
  : function
  | function_list function
  ;

function
  : type _ID
      {
        fun_idx = insert_symbol($2, FUN, $1, NO_ATTR);

        code("\n%s:", $2);
        code("\n\t\t\tPUSH\t%%14");
        code("\n\t\t\tMOV \t%%15,%%14");
      }
    _LPAREN parameter _RPAREN
      {
        set_attr(fun_idx, $5);
        var_num = 0;
      }
    body
      {
        if (get_last_element() > fun_idx)
          clear_symbols(fun_idx + 1);

        gen_sslab($2,"_exit");
        code("\n\t\t\tMOV \t%%14,%%15");
        code("\n\t\t\tPOP \t%%14");
        code("\n\t\t\tRET");
      }
  ;

type
  : _TYPE
      {  $$ = $1; }
  ;

parameter
  : /* empty */
      { $$ = 0; }

  | type _ID
      {
        insert_symbol($2, PAR, $1, 1);
        set_ptyp(fun_idx, $1);
        $$ = 1; // samo 1 parametar
      }
  ;

body
  : _LBRACKET variable_list
      {
        if (var_num)
          code("\n\t\t\tSUBS\t%%15,$%d,%%15", 4*var_num);
        gen_sslab(get_name(fun_idx), "_body");
      }
    statement_list _RBRACKET
  ;

variable_list
  : /* empty */
  | variable_list variable
  ;

variable
  : type _ID _SEMICOLON
      {
        if(lookup_symbol($2, PAR) != -1)
           err("redefinition of '%s'", $2);
        else 
           insert_symbol($2, VAR, $1, ++var_num);
      }
  ;

statement_list
  : /* empty */
  | statement_list statement
  ;

statement
  : compound_statement
  | assignment_statement
  | if_statement
  | return_statement
  ;

compound_statement
  : _LBRACKET statement_list _RBRACKET
  ;

assignment_statement
  : _ID _ASSIGN num_exp _SEMICOLON
      {
        int i;
        if( (i = lookup_symbol($1, (VAR|PAR))) == -1 )
          err("invalid lvalue '%s' in assignment", $1);
        else
          if(!check_types(i, $3))
            err("incompatible types in assignment");
        gen_mov($3, i);
      }
  ;

num_exp
  : exp

  | num_exp _AROP exp
      {
        if(!check_types($1, $3))
          err("invalid operands: arithmetic operation");
        $$ = gen_arith($2, $1, $3);
      }
  ;

exp
  : literal

  | _ID
      {
        if( ($$ = lookup_symbol($1, (VAR|PAR))) == -1)
          err("'%s' undeclared", $1);
      }

  | function_call
      {
        $$ = take_reg();
        gen_mov(FUN_REG, $$);
      }
  
  | _LPAREN num_exp _RPAREN
      { $$ = $2; }
  ;

literal
  : _INT_NUMBER
      { $$ = insert_literal($1, INT); }

  | _UINT_NUMBER
      { $$ = insert_literal($1, UINT); }
  ;

function_call
  : _ID 
      {
        if((fcall_idx = lookup_symbol($1, FUN)) == -1)
          err("'%s' is not a function", $1);
      }
    _LPAREN argument _RPAREN
      {
        if (get_attr(fcall_idx) != $4)
          err("wrong number of arguments to function '%s'",
            get_name(fcall_idx));
        gen_fcall(fcall_idx);
        gen_clear_args($4);
        set_reg_type(FUN_REG, get_type(fcall_idx));
        $$ = FUN_REG;
      }
  ;

argument
  : /* empty */
      { $$ = 0; }

  | num_exp
      { 
        if(get_ptyp(fcall_idx) != get_type($1))
           err("incompatible type for argument in '%s'",
              get_name(fcall_idx));
        gen_push($1);
        $$ = 1; // samo 1 argument
      }
  ;

if_statement
  : if_part %prec ONLY_IF
      { gen_snlab("exit", lab_num); }

  | if_part _ELSE statement
      { gen_snlab("exit", lab_num);}
  ;

if_part
  : _IF _LPAREN
      {
        lab_num++;
        gen_snlab("if", lab_num);
      }
    rel_exp
      {
        code("\n\t\t\t%s\t@false%d", 
             get_opjump($4),lab_num);
        gen_snlab("true", lab_num);
      }
    _RPAREN statement
      {
        code("\n\t\t\tJMP \t@exit%d", lab_num);
        gen_snlab("false", lab_num);
      }
  ;

rel_exp
  : num_exp _RELOP num_exp
      {
        if(!check_types($1, $3))
          err("invalid operands to relational operator");
        $$ = $2 + (get_type($1) - 1) * RELOP_NUMBER;
        gen_cmp($1, $3);
      }
  ;

return_statement
  : _RETURN num_exp _SEMICOLON
      {
        if(!check_types(fun_idx, $2))
          err("incompatible types in return");
        gen_mov($2, FUN_REG);
        code("\n\t\t\tJMP \t@%s_exit",
             get_name(fun_idx));        
      }
  ;

%%

int yyerror(char *s) {
  fprintf(stderr, "\nline %d: ERROR: %s\n", yylineno, s);
  error_count++;
  return 0;
}

void warning(char *s) {
  fprintf(stderr, "\nline %d: WARNING: %s", yylineno, s);
  warning_count++;
}

// Proverava tipove 2 elementa u tabeli simbola.
bool check_types(int first_index, int second_index) {
   unsigned t1 = get_type(first_index);
   unsigned t2 = get_type(second_index);
   if(t1 == t2 && t1 != NO_TYPE)
      return TRUE;
   else
      return FALSE;
}

int main() {
  int synerr;
  init_symtab();
  output = fopen("output.asm", "w+");

  synerr = yyparse();

  clear_symtab();
  fclose(output);
  
  if(warning_count)
    printf("\n%d warning(s).\n", warning_count);

  if(error_count) {
    remove("output.asm");
    printf("\n%d error(s).\n", error_count);
  }

  if (synerr)
    return -1;
  else
    return error_count;
}

